package ru.dbotthepony.mc.otm.graph.storage

import net.minecraft.world.level.Level
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.capability.energy.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.storage.IStorage
import ru.dbotthepony.mc.otm.storage.IStorageStack
import ru.dbotthepony.mc.otm.storage.StorageStackType
import java.util.*

open class BasicStorageGraphNode(private val energyDemander: IMatteryEnergyStorage? = null) : IStorageGraphNode {
	private var resolver = LazyOptional.of<IStorageGraphNode> { this }
	private var valid = true
	protected val components = ArrayList<IStorage<*>>()

	/**
	 * Setting this to true will render non functional default [attachComponents],
	 * make [addStorageComponent] and [removeStorageComponent] not add/remove components
	 * to attached graph, and [revive] not re-attach components.
	 *
	 * [invalidate] and [removeComponents] still detach all components
	 */
	protected var manualAttaching = false

	@Suppress("LeakingThis")
	final override val storageNode = Graph6Node<IStorageGraphNode>(this)

	private var demandingEnergy = false

	override fun attachComponents(to: StorageNetworkGraph) {
		if (energyDemander != null) {
			if (energyDemander.missingPower.isPositive) {
				demandingEnergy = true
				to.powerDemandingNodes.add(energyDemander)
			}
		}

		if (manualAttaching) {
			return
		}

		for (component in components) {
			to.add(component)
		}
	}

	fun tickEnergyDemanding() {
		energyDemander ?: throw IllegalStateException("No energy demander")

		if (!demandingEnergy && storageGraph != null && energyDemander.missingPower.isPositive) {
			demandingEnergy = true
			storageGraph!!.powerDemandingNodes.add(energyDemander)
		} else if (demandingEnergy && storageGraph != null && !energyDemander.missingPower.isPositive) {
			demandingEnergy = false
			storageGraph!!.powerDemandingNodes.remove(energyDemander)
		}
	}

	override fun removeComponents(from: StorageNetworkGraph) {
		for (component in components) {
			from.remove(component)
		}

		if (energyDemander != null) {
			from.powerDemandingNodes.remove(energyDemander)
			demandingEnergy = false
		}
	}

	@Suppress("unchecked_cast")
	fun <T : IStorageStack, U : IStorage<T>> computeIfAbsent(identity: StorageStackType<T>, provider: (StorageStackType<T>) -> U): U {
		for (component in components) {
			if (component.storageType === identity) {
				return component as U
			}
		}

		val factory = provider(identity)
		addStorageComponent(factory)
		return factory
	}

	fun addStorageComponent(component: IStorage<*>) {
		for (component1 in components) {
			if (component === component1 || component1.storageType === component.storageType) {
				return
			}
		}

		components.add(component)

		if (valid && !manualAttaching)
			storageGraph?.add(component)
	}

	fun removeStorageComponent(component: IStorage<*>) {
		var indexOf = -1

		for ((i, component1) in components.withIndex()) {
			if (component === component1 || component1.storageType === component.storageType) {
				indexOf = i
				break
			}
		}

		if (indexOf == -1) {
			return
		}

		val self = components[indexOf]
		components.removeAt(indexOf)

		if (valid)
			storageGraph?.remove(self)
	}

	fun removeStorageComponent(component: StorageStackType<*>) {
		var indexOf = -1

		for ((i, component1) in components.withIndex()) {
			if (component1.storageType === component) {
				indexOf = i
				break
			}
		}

		if (indexOf == -1) {
			return
		}

		val self = components[indexOf]
		components.removeAt(indexOf)

		if (valid && !manualAttaching)
			storageGraph?.remove(self)
	}

	open fun invalidate() {
		if (!valid) return
		valid = false
		resolver.invalidate()

		val storageGraph = storageGraph

		if (storageGraph != null) {
			for (component in components) {
				storageGraph.remove(component)
			}
		}
	}

	open fun revive() {
		if (valid) return
		valid = true
		resolver = LazyOptional.of { this }

		val storageGraph = storageGraph

		if (storageGraph != null && !manualAttaching) {
			for (component in components) {
				storageGraph.add(component)
			}
		}
	}

	fun get(): LazyOptional<IStorageGraphNode> {
		return if (valid) resolver else LazyOptional.empty()
	}

	fun destroy(level: Level?) {
		if (level != null) {
			storageNode.destroy { StorageNetworkGraph(level) }
		} else {
			storageGraph?.removeNode(storageNode)
		}
	}
}
